/*--------------------------------------------------------------------------------- */
/* DOS.H - All needed structs, defines, etc used for communication between          */
/*         DOS and Windows (including Glide, dgVesa)                                */
/*                                                                                  */
/* Copyright (C) 2003 Dege                                                          */
/*                                                                                  */
/* This library is free software; you can redistribute it and/or                    */
/* modify it under the terms of the GNU Lesser General Public                       */
/* License as published by the Free Software Foundation; either                     */
/* version 2.1 of the License, or (at your option) any later version.               */
/*                                                                                  */
/* This library is distributed in the hope that it will be useful,                  */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of                   */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU                */
/* Lesser General Public License for more details.                                  */
/*                                                                                  */
/* You should have received a copy of the GNU Lesser General Public                 */
/* License along with this library; if not, write to the Free Software              */
/* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA   */
/*--------------------------------------------------------------------------------- */

/*------------------------------------------------------------------------------------------*/
/* dgVoodoo: Dos.h																			*/
/*			 DOS-kommunikci																*/
/*------------------------------------------------------------------------------------------*/

#ifndef	DOS_H
#define DOS_H

#include "dgVoodooConfig.h"

#ifdef __MSC__
#include "vddsvc.h"
#endif

/* Ezt a header-t a Dos-driver fordtsakor is hasznljuk, ezrt van az albbi definci */
#ifndef __MSC__
#define HANDLE	unsigned int
#define HWND	unsigned int
#endif

/*------------------------------------------------------------------------------------------*/
/*...................................... Defincik ........................................*/

/* Az egyes platformok kdjai */
#ifndef	PLATFORM_CODES
#define PLATFORM_CODES

#define	PLATFORM_WINDOWS	0
#define	PLATFORM_DOSWIN9X	1
#define PLATFORM_DOSWINNT	2


/* A szervernek kldtt zenetek */
#define	DGSM_PROCEXECBUFF		(WM_USER+0)	/* A kzs terleten lev parancsok (fggvnyek) vgrehajtsa */
#define	DGSM_SETCONSOLEHWND		(WM_USER+1) /* A Dos-os prg konzolablaknak HWND-jnek eltrolsa */
#define	DGSM_SETTIMER			(WM_USER+2)
#define	DGSM_DGVOODOORESET		(WM_USER+3)
#define	DGSM_SETCLIENTHWND		(WM_USER+4)
#define	DGSM_GETVESAMODEFORMAT	(WM_USER+5)
#define DGSM_CLOSEBINDING		(WM_USER+6)

/* A kliensnek kldtt zenetek */
#define	DGCM_VESAGRABACTBANK	(WM_USER+64)
#define	DGCM_SUSPENDDOSTHREAD	(WM_USER+65)
#define	DGCM_RESUMEDOSTHREAD	(WM_USER+66)
#define DGCM_MOUSEDRIVERSTRUC	(WM_USER+1)
#define DGCM_MOUSEINTERRUPT		(WM_USER+68)


#endif

/*------------------------------------------------------------------------------------------*/
/*..................................... Struktrk .........................................*/

/* Az aktulis videmd s annak pixelformtuma (VESA-hoz) */
typedef struct	{

	unsigned short	ModeNumber;
	unsigned short	XRes;
	unsigned short	YRes;
	unsigned char	BitPP;
	unsigned char	RedSize;
	unsigned char	RedPos;
	unsigned char	GreenSize;
	unsigned char	GreenPos;
	unsigned char	BlueSize;
	unsigned char	BluePos;
	unsigned char	RsvSize;
	unsigned char	RsvPos;
	unsigned char	reserved;

} dgVoodooModeInfo;


/* Egrkurzor-inf: jelenleg hasznlaton kvli */
typedef struct	{
	unsigned int	x;
	unsigned int	y;
	unsigned int	brecttop_x, brecttop_y;
	unsigned int	brectbottom_x, brectbottom_y;
} MouseInfo;


/* Regiszterstruktra a VESA paramterekhez */
typedef struct	{

	union {
		struct {
			unsigned char	al;
			unsigned char	ah;
		};
		unsigned short	ax;
		unsigned long	eax;
	};
	union {
		struct {
			unsigned char	bl;
			unsigned char	bh;
		};
		unsigned short	bx;
		unsigned long	ebx;
	};
	union {
		struct {
			unsigned char	cl;
			unsigned char	ch;
		};
		unsigned short	cx;
		unsigned long	ecx;
	};
	union {
		struct {
			unsigned char	dl;
			unsigned char	dh;
		};
		unsigned short	dx;
		unsigned long	edx;
	};
	union {
		unsigned short	si;
		unsigned long	esi;
	};
	union {
		unsigned short	di;
		unsigned long	edi;
	};
	unsigned long	es;

} VESARegisters;


/*------------------------------------------------------------------------------------------*/

/* A kzs memriaterlet tpusa (regisztrcis vagy kommunikcis terlet). */
/* A regisztrcis struktra akkor hasznlatos, amikor a wrapper XP alatt szerver mdban fut. */
#define AREAID_REGSTRUCTURE_GLIDE211	2
#define AREAID_REGSTRUCTURE_GLIDE243	3
#define AREAID_COMMAREA					1


#define	LFBMAXPIXELSIZE					4

#ifndef GLIDE1

#define AREAID_REGSTRUCTURE				AREAID_REGSTRUCTURE_GLIDE243

#else

#define AREAID_REGSTRUCTURE				AREAID_REGSTRUCTURE_GLIDE211

#endif

typedef struct	{

	unsigned int	areaId;					/* A kzs memriaterlet tpusa */

} AreaIdPrefix;


/*------------------------------------------------------------------------------------------*/

/* A fggvnykdoknak s az tadand adatoknak fenntartott terletek mrete */
#define EXEBUFFSIZE		((32*1024)/4)
#define FUNCDATASIZE	1536*1024


/* Flagek a kernelflag mezhz */
/* (Ezek olyan flagek, amelyeket DOS-bl, s kernelmdban/VDD-ben egyarnt hasznlunk */

#define KFLAG_INKERNEL			0x1			/* A wrapper ppen kernelmdban dolgozik */
#define KFLAG_VMWINDOWED		0x2			/* Win9x/Me: a kiszolglt VM ablakos mdban indtotta a szervert - nem hasznlt*/
#define KFLAG_SCREENSHOT		0x4			/* A megfelel billenty a screenshot ksztshez lenyomva */
#define KFLAG_GLIDEACTIVE		0x8			/* A glide aktv - nem hasznlt */
#define KFLAG_SERVERWORKING		0x20		/* A szerver a megadott fggvnyeket dolgozza fel */
#define KFLAG_CTRLALT			0x80		/* Win9x/Me: a Ctrl-Alt lenyomva, egrfkusz-elengedsi rutint meghvni */
#define KFLAG_PALETTECHANGED	0x100		/* A DOS-os VM megvltoztatta a VGA/VESA palettt */
#define KFLAG_VESAFRESHDISABLED	0x200		/* A VESA-kpfrissts tiltva */
#define KFLAG_DRAWMUTEXUSED		0x400		/* A rajzolsi mutexet birtokolja a DOS-os szl */
#define KFLAG_SUSPENDDOSTHREAD	0x800		/* A rajzolsi mutex elengedse utn a DOS-szlat suspendelni kell */


/* A kommunikcis terlet felptse (ez van az osztott memriaterleten) */
typedef struct	{
	
	AreaIdPrefix	areaId;					/* Terlet (struktra) tpusa: AREAID_COMMAREA */
	unsigned int	kernelflag;				/* klnbz flagek a DOS-os s Windowsos rsz kzti kommunikcihoz */
	unsigned int	ExeCodeIndex;			/* Eltrolt fggvnykdok indexe */
	unsigned char	*FDPtr;					/* Mutat az tadand adatokhoz (a FuncData mezn bell): szabad terlet */
	unsigned char	progname[128];			/* Dos-os program neve */
	union {
		unsigned int	WinOldApHWND;			/* A Dos-os prg konzolablaknak HWND-je (csak Win9x/Me) */
		HWND			consoleHwnd;	
	};
	unsigned int	VESAPalette[256];		/* VESA: aktulis paletta, ARGB */
	unsigned char	VESASession;			/* VESA: 1=aktv, 0=inaktv */
	unsigned char	VGAStateRegister3DA;	/* virtulis VGA 3DA (vertical blank) regisztere */
	unsigned char	VGARAMDACSize;			/* virtulis VGA RAMDAC mrete: 6 vagy 8 bit */
	unsigned char	Reserved;
	void			*VideoMemory;			/* virtulis vidememria lineris cme */
	void			*videoMemoryClient;		/* virtulis vidememria lineris cme a kliens cmterben */
	unsigned int	BytesPerScanLine;		/* VESA: bytes per scan line */
	unsigned int	DisplayOffset;			/* VESA: display offset */
	dgVoodooModeInfo actmodeinfo;			/* Az aktulis VESA md infja */
	
    unsigned long	ExeCodes[EXEBUFFSIZE];	/* Fggvnykdoknak fenntartott hely */
	unsigned char	FuncData[FUNCDATASIZE];	/* tadand adatoknak fenntartott hely */
	unsigned char	LFB[];					/* A struktra kibvtve az LFB szmra fenntartott hellyel */
											/* Csak WinXp szervermdban */

} CommArea;

/*------------------------------------------------------------------------------------------*/
/* Adatok, amelyeket a szerver tad a kernelmodulnak, amikor regisztrlja magt				*/
typedef struct	{

	AreaIdPrefix	areaId;					/* A kzs memriaterlet tpusa: AREAID_REGSTRUCTURE */
	HANDLE			hidWinHwnd;				/* A dolgoz szl rejtett ablaknak kezelje */
	CommArea		*serveraddrspace;		/* A kzs terlet cme a szerver cmterben */
	dgVoodooConfig  serverConfig;			/* A szerver konfigja */

} ServerRegInfo;


/*------------------------------------------------------------------------------------------*/
/* A kliens ltal hvott szolgltatsok (a DOS-os overlay-bl)								*/
/* (Win98/Me alatt a VXD dolgozza fel ket, NT alatt kzvetlenl a szerver)					*/
#define DGCLIENT_BEGINGLIDE				0x1		/* Init s a kliens regisztrlsa */
#define DGCLIENT_ENDGLIDE				0x2		/* kliens eldobsa, cleanup */
#define DGCLIENT_CALLSERVER				0x3		/* Szerver meghvsa, a pufferelt fggvnyek*/
												/* vgrehajtsa */
#define DGCLIENT_GETPLATFORMTYPE		0x4		/* Platform lekrdezse */
#define DGCLIENT_RELEASE_TIME_SLICE		0x5		/* Idszelet elengedse (csak NT) */
#define DGCLIENT_RELEASE_MUTEX			0x6		/* Rajzolsi mutex elengedse (csak NT) */
#define DGCLIENT_VESA					0x7
#define DGCLIENT_BEGINVESA				0x8
#define DGCLIENT_ENDVESA				0x9
#define DGCLIENT_INSTALLMOUSE			0xA
#define DGCLIENT_UNINSTALLMOUSE			0xB
#define DGCLIENT_GETVERSION				0xC

/*------------------------------------------------------------------------------------------*/
/* Az egyes Glide-fggvnyek kdjai */

#define GRGLIDEINIT						0
#define	GRSSTWINOPEN					1
#define GRSSTWINCLOSE					2
#define GRGLIDESHUTDOWN					3

#define GRBUFFERCLEAR					4
#define GRBUFFERSWAP					5

#define GUDRAWTRIANGLEWITHCLIP			6
#define GUAADRAWTRIANGLEWITHCLIP		7
#define GUDRAWPOLYGONVERTEXLISTWITHCLIP 8
#define GRDRAWTRIANGLE					9
#define GRDRAWPOLYGONVERTEXLIST			10
#define GRAADRAWPOLYGON					11
#define GRAADRAWPOLYGONVERTEXLIST		12
#define GRAADRAWTRIANGLE				13
#define GRDRAWPLANARPOLYGON				14
#define GRDRAWPLANARPOLYGONVERTEXLIST	15
#define GRDRAWPOLYGON					16
#define GRAADRAWLINE					17
#define GRDRAWLINE						18
#define GRAADRAWPOINT					19
#define GRDRAWPOINT						20

#define GRTEXDOWNLOADMIPMAP				21
#define GRTEXDOWNLOADMIPMAPLEVEL		22
#define GRTEXDOWNLOADMIPMAPLEVELPARTIAL	23
#define GRTEXDOWNLOADTABLE				24
#define GRTEXDOWNLOADTABLEPARTIAL		25
#define GRTEXNCCTABLE					26
#define GRTEXSOURCE						27
#define GRTEXMIPMAPMODE					28
#define GRTEXCOMBINE					29
#define GRTEXFILTERMODE					30
#define GRTEXCLAMPMODE					31
#define GRTEXLODBIASVALUE				32
#define GRTEXMULTIBASE					33
#define GRTEXMULTIBASEADDRESS			34

#define GUTEXALLOCATEMEMORY				35
#define GUTEXCHANGEATTRIBUTES			36
#define GUTEXCOMBINEFUNCTION			37
#define GUTEXDOWNLOADMIPMAP				38
#define GUTEXDOWNLOADMIPMAPLEVEL		39
#define GUTEXGETCURRENTMIPMAP			40
#define GUTEXGETMIPMAPINFO				41
#define GUTEXMEMQUERYAVAIL				42
#define GUTEXMEMRESET					43
#define GUTEXSOURCE						44
#define GU3DFGETINFO					45
#define GU3DFLOAD						46
#define GLIDEGETUTEXTURESIZE			47

#define GRCOLORCOMBINE					48
#define GUCOLORCOMBINEFUNCTION			49
#define GRCONSTANTCOLORVALUE4			50
#define GRCOLORMASK						51
#define GRCONSTANTCOLORVALUE			52

#define GRALPHABLENDFUNCTION			53
#define GRALPHACOMBINE					54
#define GUALPHASOURCE					55
#define GRALPHATESTFUNCTION				56
#define GRALPHATESTREFERENCEVALUE		57

#define GRFOGCOLORVALUE					58
#define GRFOGMODE						59
#define GRFOGTABLE						60
#define GLIDEGETINDTOWTABLE				61
#define GUFOGTABLEINDEXTOW				62
#define GUFOGGENERATEEXP				63
#define GUFOGGENERATEEXP2				64
#define GUFOGGENERATELINEAR				65

#define GRRENDERBUFFER					66
#define GRSSTORIGIN						67
#define GRCLIPWINDOW					68
#define GRCULLMODE						69
#define GRDISABLEALLEFFECTS				70
#define GRDITHERMODE					71
#define GRGAMMACORRECTIONVALUE			72
#define GRGLIDEGETSTATE					73
#define GRGLIDESETSTATE					74
#define GRHINTS							75

#define GRSSTSCREENWIDTH				76
#define GRSSTSCREENHEIGHT				77
#define GRSSTSTATUS						78
#define GRSSTVIDEOLINE					79
#define GRSSTVRETRACEON					80
#define GRSSTCONTROLMODE				81
#define GRBUFFERNUMPENDING				82
#define GRRESETTRISTATS					83
#define GRTRISTATS						84

#define GRDEPTHBUFFERMODE				85
#define GRDEPTHBUFFERFUNCTION			86
#define GRDEPTHMASK						87
#define GRLFBCONSTANTALPHA				88
#define GRLFBREADREGION					89
#define GRLFBWRITEREGION				90
#define GRDEPTHBIASLEVEL				91
#define GLIDEGETCONVBUFFXRES			92

#define GRCHROMAKEYMODE					93
#define GRCHROMAKEYVALUE				94

#define GRLFBLOCK						95
#define GRLFBUNLOCK						96
#define GRLFBWRITECOLORFORMAT			97
#define GRLFBWRITECOLORSWIZZLE			98
#define GLIDESETUPLFBDOSBUFFERS			99

/* Glide1-fggvnyek */
#define GRSSTOPEN						100
#define GRLFBBYPASSMODE					101
#define GRLFBORIGIN						102
#define GRLFBWRITEMODE					103
#define GRLFBBEGIN						104
#define GRLFBEND						105
#define GRLFBGETREADPTR					106
#define GRLFBGETWRITEPTR				107
#define GUFBREADREGION					108
#define GUFBWRITEREGION					109

/* VESA-fggvnyek */
#define VESASETVBEMODE					110
#define VESAUNSETVBEMODE				111
#define VESAFRESH						112

/* Ezeket a fggvnyeket a kernelmodul hivja */
#define DGVOODOOCLIENTCRASHED			113
#define DGVOODOONEWCLIENTREGISTERING	114

/* Egyb fggvnyek */
#define DGVOODOOGETCONFIG				115
#define DGVOODOORELEASEFOCUS			116

#define GLIDEINSTALLMOUSE				117
#define GLIDEUNINSTALLMOUSE				118
#define GLIDEISMOUSEINSTALLED			119

#define VESAGETXRES						120
#define VESAGETYRES						121

#define PORTLOG_IN						122
#define PORTLOG_OUT						123


/* Ez a kd zrja le a kdszekvencit */
#define GLIDEENDITEM					0xFFFFFFFF

#ifdef __MSC__
/*------------------------------------------------------------------------------------------*/
/*............................. Bels fggvnyek predeklarcija ...........................*/

void					HideMouseCursor(int);
void					RestoreMouseCursor(int);
void					CreateRenderingWindow();
void					DestroyRenderingWindow();
void					CreateServerCommandWindow();
void					DestroyServerCommandWindow();
void					SetWindowClientArea(HWND, int, int);
int						IsWindowSizeIsEqualToResolution (HWND window);
void					AdjustAspectRatio (HWND window, int x, int y);
unsigned int			GetIdealWindowSizeForResolution (unsigned int xRes, unsigned int yRes);
unsigned char*			DosGetSharedLFB();
void					DosSendSetTimerMessage();
DWORD					NTCallFunctionOnServer (unsigned long funcCode, unsigned int pNum, unsigned long *params);
int						DOSInstallIOHook (unsigned short firstPort, unsigned short lastPort,
										  void (_stdcall *inHandler)(unsigned short, unsigned char *),
										  void (_stdcall *outHandler)(unsigned short, unsigned char) );
void					DOSUninstallIOHook (unsigned short firstPort, unsigned short lastPort);
void					ExitDosNt ();
int						InitVDDRendering ();
void					ShutDownVDDRendering ();
int						OpenServerCommunicationArea (ServerRegInfo *regInfo);
void					CloseServerCommunicationArea ();
void*					DosMapFlat (unsigned short selector, unsigned int offset);

/*------------------------------------------------------------------------------------------*/
/*.................................. Globlis vltozk .....................................*/
extern	CommArea		*c;
extern	HWND			warningBoxHwnd;
extern	HWND			renderingWinHwnd;
extern	HWND			serverCmdHwnd;
extern	HWND			clientCmdHwnd;
extern	HWND			consoleHwnd;
extern	unsigned int	actResolution;
extern	void			(_stdcall *_call_ica_hw_interrupt)(int ms, int line, int count);
extern	HANDLE			hDevice;
extern	ServerRegInfo	regInfo;

#endif

#endif	/* DOS_H */