/*--------------------------------------------------------------------------------- */
/* dgVoodooConfig.h - Config structure and flags bits                               */
/*                                                                                  */
/* Copyright (C) 2004 Dege                                                          */
/*                                                                                  */
/* This library is free software; you can redistribute it and/or                    */
/* modify it under the terms of the GNU Lesser General Public                       */
/* License as published by the Free Software Foundation; either                     */
/* version 2.1 of the License, or (at your option) any later version.               */
/*                                                                                  */
/* This library is distributed in the hope that it will be useful,                  */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of                   */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU                */
/* Lesser General Public License for more details.                                  */
/*                                                                                  */
/* You should have received a copy of the GNU Lesser General Public                 */
/* License along with this library; if not, write to the Free Software              */
/* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA   */
/*--------------------------------------------------------------------------------- */


/*------------------------------------------------------------------------------------------*/
/* dgVoodoo: dgVoodooConfig.h																*/
/*			 Konfigurcihoz szksges defincik											*/
/*------------------------------------------------------------------------------------------*/

#ifndef		DGVOODOOCONFIG_H
#define		DGVOODOOCONFIG_H

/*------------------------------------------------------------------------------------------*/
/* Setup: azonost a kirt konfig-fjlokban (.cfg) */
#define		CFGFILEID	"DGVOODOOCFG"

/*------------------------------------------------------------------------------------------*/
/*.................................. Defincik a konfighoz ................................*/

/* Alaprtelmezett rtkek a config struktra mezihez */
#define	TEXTUREMEMSIZE				4*1024*1024			/* Voodoo-textramem mrete */
#define	CFG_DEFFLAGS				0
#define	CFG_DEFFLAGS2				0
#define DEFAULT_DOS_TIMER_BOOST		80

/* Konstansok a konfig struktra flagjeihez */
#define CFG_WBUFFER					0x1					/* W-buffert mindig Z-bufferen keresztl emull */
#define CFG_RELYONDRIVERFLAGS		CFG_WBUFFER
#define CFG_WINDOWED				0x2					/* Ablakos md */
#define CFG_DISPMODE32BIT			0x4					/* 32 bites videomd */
#define CFG_SETREFRESHRATE			0x8					/* A frisstsi freq-t az alkalmazs */
														/* lltja be (egybknt az akt. freq) */
#define CFG_PERFECTTEXMEM			0x10				/* Tkletes textramemria-emulci */
#define CFG_WBUFFERMETHODISFORCED	0x80				/* A W-bufferels mdszere knyszertett:*/
														/*	- CFG_WBUFFER flag set: mindig valdi w-bufferels */
														/*	- CFG_WBUFFER flag not set: mindig emullt w-bufferels */
														/* vagy a W-bufferels mdszere detektlt, ekkor */
														/*  - CFG_RELYONDRIVERFLAGS set: a driver ltal a W-bufferels tmogatsra */
														/*								 visszaadott flag alapjn dnt */
														/*  - CFG_RELYONDRIVERFLAGS not set: teszt alapjn dnt */
#define CFG_VESAENABLE				0x100				/* Beptett VESA-tmogats engedlyezve */
#define CFG_SETMOUSEFOCUS			0x200               /* Egrfkusz az alkalmazsnl */
#define CFG_SUPPORTMODE13H			0x400               /* 320x200x256 VGA-md tmogatsa (DOS, VESA emu) */
#define CFG_TEXTURES32BIT			0x800				/* mindig 32 bites textrk */
#define CFG_TEXTURES16BIT			0x1000				/* mindig 16 bites textrk */
#define CFG_STOREMIPMAP				0x2000				/* Mipmapek msolatnak eltrolsa */
#define CFG_GRABENABLE				0x4000              /* Kpernylops engedlyezve */
#define CFG_GRABCLIPBOARD			0x8000              /* Kplops a vglapra */
#define CFG_CKMFORCETNTINAUTOMODE	0x10000
#define CFG_TOMBRAIDER				0x20000             /* TR1 rnykhiba-javts */
#define CFG_HUNGARIAN				0x40000             /* A nyelv magyar */
#define CFG_HWLFBCACHING			0x80000             /* Hardveres gyorstpufferek az LFB-hez */
#define CFG_LFBREALVOODOO			0x100000            /* Kzelebb az igazi hardverhez */
#define CFG_TRIPLEBUFFER			0x200000            /* Mindig triplapufferels */
#define CFG_ACTIVEBACKGROUND		0x400000			/* A httrben is aktv a DOS-os prg */
#define CFG_LFBFASTUNMATCHWRITE		0x800000			/* Gyorsrs a nem egyez LFB formtumokhoz */
#define CFG_CTRLALT					0x1000000           /* Egrfkusz elengedse Ctrl-Alt-ra */
#define CFG_VSYNC					0x2000000           /* Fggleges visszafutst mindig megvrni */
#define CFG_NTVDDMODE				0x4000000           /* a wrapper VDD-mdban mkdik (XP-hez) */
#define CFG_DISABLEMIPMAPPING		0x8000000           /* Mipmapping tiltsa */
#define CFG_ENABLEGAMMARAMP			0x10000000          /* Glide-gammaramp engedlyezse */
#define CFG_ENABLEHWVBUFFER			0x20000000          /* Hardveres vertex pufferek hasznlata */
#define CFG_BILINEARFILTER			0x40000000          /* Mindig bilineris szrs a textrkhoz */
#define CFG_CLIPPINGBYD3D			0x80000000          /* Vgs a Direct3D-n keresztl */

#define CFG_CLOSESTREFRESHFREQ		0x1					/* Frisstsi frekvencia: az alk. ltal ignyelthez legkzelebb es */
#define CFG_FORCECLIPOPF            0x2                 /* CLI/POPF megfelel kezelse - rejtett opci */
#define CFG_LFBRESCALECHANGESONLY	0x4					/* eltr felbontsok esetn csak a vltozsok tmretezse, */
														/* s nem az egsz kp (gyorsrst ignyel) */
#define CFG_LFBFORCEFASTWRITE		0x8					/* Gyorsrs hasznlata egyez formtumokhoz is - rejtett opci */
#define CFG_YMIRROR					0x10				/* Kperny fggleges tkrzse (orig-tesztekhez) - rejtett opci */
#define CFG_FORCETRILINEARMIPMAPS	0x20				/* Trilineris mipmapping knyszertse */
#define CFG_PREFERTMUWIFPOSSIBLE	0x40				/* tbls kd, z-buffer s TMUDIFFW hint esetn a tmu w-t */
														/* hasznlja w-knt a kdhz is, s nem a valdi w-t */
														/* (ezltal a kd perspektvja romlik el, nem a textr) */
#define CFG_LFBALWAYSFLUSHWRITES	0x80				/* A rgi lfb-unlock dude-mdja: mindig kirja a vltozsokat
															(nem cache-el) (kompatibilitsi s teszt-opci, rejtett) */
#define CFG_DEPTHEQUALTOCOLOR		0x100				/* A kppuffer s a mlysgi puffer bitmlysge legyen mindig egyenl */
														/* (kompatibilitsi s teszt-opci, rejtett) */
#define CFG_DISABLELFBREAD			0x200				/* LFB elrsnek tiltsa olvassra */
#define CFG_DISABLELFBWRITE			0x400				/* LFB elrsnek tiltsa rsra */
#define CFG_AUTOGENERATEMIPMAPS		0x800				/* A hinyz mipmap-szintek automatikus generlsa */
#define CFG_LFBNOMATCHINGFORMATS	0x1000				/* Az LFB-formtumok mindig klnbznek tekintendk - rejtett opci */
#define CFG_ALWAYSINDEXEDPRIMS		0x2000
#define CFG_HIDECONSOLE				0x4000				/* Konzol elrejtse (DOS) */
#define CFG_LFBDISABLETEXTURING		0x8000
#define CFG_CFORMATAFFECTLFB		0x10000
#define CFG_DISABLEVESAFLATLFB		0x20000
#define CFG_PRESERVEWINDOWSIZE		0x40000
#define CFG_PRESERVEASPECTRATIO		0x80000

/* A konfig flagekben lev bitmezk rtkei */
#define CFG_CKMAUTOMATIC			0x00
#define CFG_CKMALPHABASED			0x01
#define CFG_CKMNATIVE				0x02
#define CFG_CKMNATIVETNT			0x03

#define GetCKMethodValue(flags)				((flags >> 5) & 0x3)
#define SetCKMethodValue(flags, method)		flags = (flags & ~(0x3 << 5)) | (method << 5)

#define CFG_WDETFORCEZ				(CFG_WBUFFERMETHODISFORCED)
#define CFG_WDETFORCEW				(CFG_WBUFFERMETHODISFORCED | CFG_WBUFFER)
#define CFG_WDETONDRIVERFLAG		(CFG_RELYONDRIVERFLAGS)
#define CFG_WDETDRAWINGTEST			0

#define GetLfbAccessValue(flags)			(((flags & CFG_DISABLELFBWRITE) ? 2 : 0) | ((flags & CFG_DISABLELFBREAD) ? 1 : 0))

#define GetWBuffMethodValue(flags)			(((flags & CFG_WBUFFERMETHODISFORCED) ? 2 : 0) | ((flags & CFG_WBUFFER) ? 1 : 0))
#define SetWBuffMethodValue(flags, method)	flags = (flags & ~(CFG_WBUFFERMETHODISFORCED | CFG_WBUFFER)) | method


enum Languages	{

	English				= 0,
	Hungarian			= 1,
	numberOfLanguages	= 2
};

/*------------------------------------------------------------------------------------------*/
/*....................................... Struktrk .......................................*/


/* A wrapper konfigurcijt ler struktra */
typedef struct	{

	unsigned short	ProductVersion;
	unsigned short	GlideVersion;
	unsigned int	TexMemSize, Flags;
	unsigned int	Gamma, Version;
	unsigned int	VideoMemSize;
	unsigned int	VideoRefreshFreq;
	unsigned short	dxres, dyres;
	unsigned short	dispdev, dispdriver;
	unsigned int	RefreshRate;
	unsigned int	Flags2;
	unsigned short  dosTimerBoostPeriod;
	unsigned char	AlphaCKRefValue;
	unsigned char	language;
	unsigned char	reserved[4];

} dgVoodooConfig;


extern	dgVoodooConfig	config;


#endif