/*--------------------------------------------------------------------------------- */
/* MMConv.h - C-header for my MultiMedia Converter Lib                              */
/* (duplicated file)                                                                */
/*                                                                                  */
/* Copyright (C) 2004 Dege                                                          */
/*                                                                                  */
/* This library is free software; you can redistribute it and/or                    */
/* modify it under the terms of the GNU Lesser General Public                       */
/* License as published by the Free Software Foundation; either                     */
/* version 2.1 of the License, or (at your option) any later version.               */
/*                                                                                  */
/* This library is distributed in the hope that it will be useful,                  */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of                   */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU                */
/* Lesser General Public License for more details.                                  */
/*                                                                                  */
/* You should have received a copy of the GNU Lesser General Public                 */
/* License along with this library; if not, write to the Free Software              */
/* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA   */
/*--------------------------------------------------------------------------------- */


/*------------------------------------------------------------------------------------------*/
/* dgVoodoo:  MultiMedia Converter (MMX-implementci)										*/
/*				Ez a modul az egyes sznformtumok kztti konverzit						*/
/*				biztostja																	*/
/*																							*/
/*				Tmogatott formtumok (tetszleges komponens-sorrenddel):					*/
/*				-  8 bit: P8																*/
/*				- 16 bit: RGB555, RGB565, ARGB1555, ARGB4444, AP88							*/
/*				- 32 bit: RGB888, ARGB8888													*/
/*																							*/
/*				Konverzi fajti:															*/
/*					-  8bit -> 16 vagy 32 bit												*/
/*					- 16bit -> 16 vagy 32 bit												*/
/*					- 16bit -> 16 vagy 32 bit maszkolt ColorKey-jel							*/
/*					- 32bit -> 16 vagy 32 bit												*/
/*					- 32bit -> 16 vagy 32 bit maszkolt ColorKey-jel							*/
/*					(A P8 s AP88 formtum nem lehet clformtum)							*/
/*------------------------------------------------------------------------------------------*/
#ifndef	MMCONVERTER_H
#define MMCONVERTER_H

#include	"DDraw.h"
#include	"D3d.h"

/*------------------------------------------------------------------------------------------*/
/*.................................... Defincik ..........................................*/

/* A konvertls fajti */
#define CONVERTTYPE_COPY			0	/* teljes msols s konvertls */
#define CONVERTTYPE_COLORKEY		1   /* csak a nem kulcsszn pixelek konvertlsa */
#define CONVERTTYPE_RAWCOPY			2	/* csak msols (egyez formtumokhoz) */
#define CONVERTTYPE_RAWCOPYCOLORKEY	3	/* csak msols kulcssznnel (egyez formtumokhoz) */


/* Az tadott paletta tpusai */
#define PALETTETYPE_UNCONVERTED		0	/* 32 bit ARGB formtum */
#define PALETTETYPE_CONVERTED		1	/* Az adott formtumra konvertlt */


/*------------------------------------------------------------------------------------------*/
/*.................................... Struktrk  .........................................*/

/* Sznformtumot ler struktra */
typedef struct	{

	unsigned int BitPerPixel;
	unsigned int RBitCount, GBitCount, BBitCount, ABitCount;
	unsigned int RPos, GPos, BPos, APos;

} _pixfmt;


/*------------------------------------------------------------------------------------------*/
/*.................................... Fggvnyek  .........................................*/

/* Az srcFormat formtum puffert dstFormat formtumv konvertlja */
void _stdcall MMCConvertAndTransferData(_pixfmt *srcFormat, _pixfmt *dstFormat,
										unsigned int colorKey, unsigned int colorKeyMask,
										unsigned int constAlpha,
										void *src, void *dst,
										unsigned int x, unsigned int y,
										int srcPitch, int dstPitch,
										unsigned int *palette, unsigned char *paletteMap,
										int convType, int palType);

/* Adott puffert colorkey sznnel tlt fel */
void _stdcall MMCFillBuffer(void *dst, unsigned int bitsPerPixel, unsigned int colorKey,
						    unsigned int x, unsigned int y,
							int dstPitch);

/* Egy ARGB-ben megadott sznt tkonvertl a megadott formtumra */
unsigned int _stdcall MMCGetPixelValueFromARGB (unsigned int color, _pixfmt *byFmt);

/* Egy adott formtum sznt tkonvertl ARGB-re */
unsigned int _stdcall MMCGetARGBFromPixelValue (unsigned int color, _pixfmt *byFmt, unsigned int constAlpha);

/* Az srcFmt formtum sznt dstFmt formtumra konvertlja */
/* Ez a kvetkezvel ekvivalens: MMCGetPixelValueFromARGB (MMCGetARGBFromPixelValue (srcColor, srcFmt, unsigned int constAlpha), dstFmt) */
unsigned int _stdcall MMCConvertPixel (unsigned int srcColor, _pixfmt *srcFmt, _pixfmt *dstFmt, unsigned int constAlpha);

/* Egy DDPIXELFORMAT sznformtum-lert konvertl _pixfmt -re */
void _stdcall MMCConvertToPixFmt(LPDDPIXELFORMAT ddpf, _pixfmt *pf, int createMissingAlpha);

/* Kt formtumot hasonlt ssze */
int _stdcall MMCIsPixfmtTheSame(_pixfmt *format1, _pixfmt *format2);

/* Egy 32 bit ARGB formtum palettt pixfmt-formtumra konvertl */
void _stdcall MMCConvARGBPaletteToPixFmt(unsigned int *src, unsigned int *dst,
										 int from, int to, _pixfmt *dstFormat);

/* Egy 32 bit ARGB formtum palettt 32 bit ABGR-re konvertl */
void _stdcall MMCConvARGBPaletteToABGR (unsigned int *src, unsigned int *dst);

/*------------------------------------------------------------------------------------------*/


#endif
